#!/usr/bin/env python
"""
This example extends the basic motion example to show how to gather/scatter
to/from the root process as if the load and displacement scheme or structural
solver only runs on a single processor
"""
import f90nml
import numpy as np
from mpi4py import MPI
from fun3d.solvers import Flow


def gather_on_root(comm, x, y, z):
    nnodes = comm.gather(x.size)

    if comm.Get_rank() == 0:
        xg = np.zeros(sum(nnodes))
        yg = np.zeros(sum(nnodes))
        zg = np.zeros(sum(nnodes))
    else:
        xg = np.zeros(0)
        yg = np.zeros(0)
        zg = np.zeros(0)

    comm.Gatherv(x, (xg, nnodes), 0)
    comm.Gatherv(y, (yg, nnodes), 0)
    comm.Gatherv(z, (zg, nnodes), 0)
    return xg, yg, zg


def scatter_from_root(comm, xg, yg, zg, nnodes):
    x = np.zeros(nnodes)
    y = np.zeros(nnodes)
    z = np.zeros(nnodes)

    nnodes_list = comm.gather(x.size)

    comm.Scatterv((xg, nnodes_list), x, root=0)
    comm.Scatterv((yg, nnodes_list), y, root=0)
    comm.Scatterv((zg, nnodes_list), z, root=0)
    return x, y, z


def get_deform_motion(step, x, y, z, fx, fy, fz):
    """
    deforming motion is parabolic sinusoidal motion in z direction
    """
    dx = np.zeros_like(x)
    dy = np.zeros_like(y)
    dz = np.zeros_like(z)

    dz = 0.1 * y**2 * np.sin(2.0 * np.pi * step / 50.0) + fz
    return dx, dy, dz


def main():
    # limited support for options typically read from command line. See list at top of Python/flow/interface.py
    command_line_options = {}

    prior_steps = 0  # if restarting set to number of steps in the restart file
    q = 1e-3  # dynamic pressure

    nml = f90nml.read("fun3d.nml")
    nsteps = nml["code_run_control"]["steps"]

    # ==========================================================================
    # Initialize the fun3d flow solver - read namelist, partion mesh, etc.
    # ==========================================================================
    comm = MPI.COMM_WORLD

    flow = Flow()
    flow.initialize(kwargs=command_line_options, comm=comm)

    # ==========================================================================
    # Pull out the surface node information for coupling
    # ==========================================================================
    ibody = 1
    num_nodes = flow.extract_surface_num(body=ibody)

    x = np.zeros(num_nodes)
    y = np.zeros(num_nodes)
    z = np.zeros(num_nodes)
    if num_nodes > 0:
        x, y, z = flow.extract_surface(num_nodes, body=ibody)

    # get global image of coordinates on root
    x_global, y_global, z_global = gather_on_root(comm, x, y, z)

    # initialize the global force array to zero
    fx_global = np.zeros_like(x_global)
    fy_global = np.zeros_like(y_global)
    fz_global = np.zeros_like(z_global)

    # ==========================================================================
    # Time step loop
    # ==========================================================================
    for step in range(prior_steps + 1, prior_steps + 1 + nsteps):
        if comm.Get_rank() == 0:
            dxg, dyg, dzg = get_deform_motion(
                step, x_global, y_global, z_global, fx_global, fy_global, fz_global
            )
        else:
            dxg = 0.0
            dyg = 0.0
            dzg = 0.0

        # send out the displacements
        dx, dy, dz = scatter_from_root(comm, dxg, dyg, dzg, num_nodes)

        if num_nodes > 0:
            flow.input_deformation(dx, dy, dz, body=ibody)

        flow.iterate()

        fx = np.zeros(num_nodes)
        fy = np.zeros(num_nodes)
        fz = np.zeros(num_nodes)
        if num_nodes > 0:
            fx, fy, fz = flow.extract_forces(num_nodes, body=ibody)
        fx_global, fy_global, fz_global = gather_on_root(comm, fx, fy, fz)

        # scale by dynamic pressure
        fx_global *= q
        fy_global *= q
        fz_global *= q

        if comm.Get_rank() == 0:
            print(
                "Total force = ",
                np.sum(fx_global),
                np.sum(fy_global),
                np.sum(fz_global),
            )

    flow.post()


if __name__ == "__main__":
    main()
