#!/usr/bin/env python
"""
This example extends the basic example by extracting the connectivity
"""
import f90nml
import numpy as np
from mpi4py import MPI
from fun3d.solvers import Flow


def get_deform_motion(step, x, y, z):
    """
    deforming motion is parabolic sinusoidal motion in z direction
    """
    dx = np.zeros_like(x)
    dy = np.zeros_like(y)
    dz = np.zeros_like(z)

    dz = 0.1 * y ** 2 * np.sin(2.0 * np.pi * step / 50.0)
    return dx, dy, dz


def get_connectivity_in_surface_node_numbering(ids, tri_conn, quad_conn):
    nfacet = tri_conn.shape[0]
    nfaceq = quad_conn.shape[0]
    inverse_ids = {}
    for i in range(ids.size):
        inverse_ids[ids[i]] = i
    surface_tri_conn = np.zeros((nfacet, 3), dtype=int)
    surface_quad_conn = np.zeros((nfaceq, 4), dtype=int)

    # Add 1 because most mesh formats are 1 based unlike python
    for i in range(nfacet):
        for j in range(3):
            surface_tri_conn[i, j] = inverse_ids[tri_conn[i, j]] + 1
    for i in range(nfaceq):
        for j in range(4):
            surface_quad_conn[i, j] = inverse_ids[quad_conn[i, j]] + 1
    return surface_tri_conn, surface_quad_conn


def collect_mesh_on_root(comm, x, y, z, ids, tri_conn, quad_conn):
    # Note: converting ids and connectivity as floats because Gatherv apparently has issues with ints
    tris = tri_conn.flatten()
    quads = quad_conn.flatten()

    nnodes = comm.gather(x.size)
    ntris = comm.gather(tris.size)
    nquads = comm.gather(quads.size)

    if comm.Get_rank() == 0:
        xg = np.zeros(sum(nnodes))
        yg = np.zeros(sum(nnodes))
        zg = np.zeros(sum(nnodes))
        idsg = np.zeros(sum(nnodes), dtype=float)
        tri_conng = np.zeros(sum(ntris), dtype=float)
        quad_conng = np.zeros(sum(nquads), dtype=float)
    else:
        xg = np.zeros(0)
        yg = np.zeros(0)
        zg = np.zeros(0)
        idsg = np.zeros(0, dtype=float)
        tri_conng = np.zeros(0, dtype=float)
        quad_conng = np.zeros(0, dtype=float)

    comm.Gatherv(x, (xg, nnodes), 0)
    comm.Gatherv(y, (yg, nnodes), 0)
    comm.Gatherv(z, (zg, nnodes), 0)
    comm.Gatherv(ids.astype(float), (idsg, nnodes), 0)
    comm.Gatherv(tris.astype(float), (tri_conng, ntris), 0)
    comm.Gatherv(quads.astype(float), (quad_conng, nquads), 0)

    # convert back to ints and reshape connectivity to n,3 or n,4
    if comm.Get_rank() == 0:
        idsg = idsg.astype(int)
        tri_conng = tri_conng.reshape((-1, 3)).astype(int)
        quad_conng = quad_conng.reshape((-1, 4)).astype(int)
    return xg, yg, zg, idsg, tri_conng, quad_conng


def write_surface_to_tecplot(x, y, z, ids, tri_conn, quad_conn):
    nnodes = x.size
    nfacet = tri_conn.shape[0]
    nfaceq = quad_conn.shape[0]
    nfaces = nfacet + nfaceq

    fh = open("funtofem_surface.dat", mode="w")
    lines = []

    # file header
    lines.append('title="surface points and l2g id for funtofem"\n')
    lines.append('variables="x","y","z","id"\n')
    lines.append(
        'zone t="body 1", i=%i, j=%i, f=fepoint,  solutiontime= 0.0000000E+00, strandid=0\n'
        % (nnodes, nfaces)
    )
    # Nodes
    for i in range(nnodes):
        lines.append(" %10.19f %10.19f %10.19f %i\n" % (x[i], y[i], z[i], ids[i]))

    # Element connectivity
    for i in range(nfacet):
        # tecplot format says to write the last index twice for tris
        lines.append(
            " %i %i %i %i\n"
            % (tri_conn[i, 0], tri_conn[i, 1], tri_conn[i, 2], tri_conn[i, 2])
        )
    for i in range(nfaceq):
        lines.append(
            " %i %i %i %i\n"
            % (quad_conn[i, 0], quad_conn[i, 1], quad_conn[i, 2], quad_conn[i, 3])
        )

    fh.writelines(lines)
    fh.close()


def main():
    # limited support for options typically read from command line. See list at top of Python/flow/interface.py
    command_line_options = {}

    # if restarting set to number of steps in the restart file
    prior_steps = 0

    nml = f90nml.read("fun3d.nml")
    nsteps = nml["code_run_control"]["steps"]

    # ==========================================================================
    # Initialize the fun3d flow solver - read namelist, partion mesh, etc.
    # ==========================================================================
    comm = MPI.COMM_WORLD

    flow = Flow()
    flow.initialize(kwargs=command_line_options, comm=comm)

    # ==========================================================================
    # Pull out the surface node information for coupling
    # ==========================================================================
    # body number for multibody problems. this demo only has 1 body.
    ibody = 1  # fun3d numbering so starts at 1

    num_nodes = flow.extract_surface_num(body=ibody)

    # get the initial coordinates
    x = np.zeros(num_nodes)
    y = np.zeros(num_nodes)
    z = np.zeros(num_nodes)
    ids = np.zeros(num_nodes, dtype=int)
    if num_nodes > 0:
        x, y, z = flow.extract_surface(num_nodes, body=ibody)

        # get the global id numbers if desired
        ids = flow.extract_surface_id(num_nodes, body=ibody)

    # Get the local mesh connectivity (based on global volume node ids)
    # only includes faces owned by this processor
    num_tris = flow.extract_tri_face_num(body=ibody)
    tri_conn = np.zeros((num_tris, 3), dtype=int)
    if num_tris > 0:
        tri_conn = flow.extract_tri_face_connectivity(num_tris, body=ibody)

    num_quads = flow.extract_quad_face_num(body=ibody)
    quad_conn = np.zeros((num_quads, 4), dtype=int)
    if num_quads > 0:
        quad_conn = flow.extract_quad_face_connectivity(num_quads, body=ibody)

    # Collect the connectivity on root rank and write to a file
    xg, yg, zg, idsg, tri_conng, quad_conng = collect_mesh_on_root(
        comm, x, y, z, ids, tri_conn, quad_conn
    )
    if comm.Get_rank() == 0:
        # Convert to connectivity based on surface node numbering since interfacing code may not know about volume mesh
        surf_tri_conng, surf_quad_conng = get_connectivity_in_surface_node_numbering(
            idsg, tri_conng, quad_conng
        )
        write_surface_to_tecplot(xg, yg, zg, idsg, surf_tri_conng, surf_quad_conng)

    # ==========================================================================
    # Time step loop
    # ==========================================================================
    for step in range(prior_steps + 1, prior_steps + 1 + nsteps):

        if num_nodes > 0:
            dx, dy, dz = get_deform_motion(step, x, y, z)
            flow.input_deformation(dx, dy, dz, body=ibody)

        flow.iterate()

        if num_nodes > 0:
            fx, fy, fz = flow.extract_forces(num_nodes, body=ibody)
        else:
            fx = 0.0
            fy = 0.0
            fz = 0.0

        fx_total = comm.allreduce(np.sum(fx))
        fy_total = comm.allreduce(np.sum(fy))
        fz_total = comm.allreduce(np.sum(fz))
        if comm.Get_rank() == 0:
            print("Total surface force", fx_total, fy_total, fz_total)

    # ==========================================================================
    # Finish the fun3d analysis. Write restart files, hist files, etc
    # ==========================================================================
    flow.post()


if __name__ == "__main__":
    main()
