#!/usr/bin/env python
"""
This example extends the basic motion example to show how to couple two deforming
bodies
"""
import f90nml
import numpy as np
from mpi4py import MPI
from fun3d.solvers import Flow


def get_deform_motion(step, x, y, z, ibody):
    """
    deforming motion is parabolic spatially and sinusoidal in time
    """
    dx = np.zeros_like(x[ibody])
    dy = np.zeros_like(y[ibody])
    dz = np.zeros_like(z[ibody])

    if ibody == 1:
        sign = 1.0
        offset = 0.0
    else:
        offset = 3.0
        sign = -1.0

    dz = (
        sign * 0.4 * (x[ibody] - offset) ** 2 * np.sin(2.0 * np.pi * float(step) / 50.0)
    )
    return dx, dy, dz


def main():
    command_line_options = {}
    nbodies = 2

    # if restarting set to number of steps in the restart file
    prior_steps = 0

    nml = f90nml.read("fun3d.nml")
    nsteps = nml["code_run_control"]["steps"]

    # ==========================================================================
    # Initialize the fun3d flow solver - read namelist, partion mesh, etc.
    # ==========================================================================
    comm = MPI.COMM_WORLD

    flow = Flow()
    flow.initialize(kwargs=command_line_options, comm=comm)

    # ==========================================================================
    # Pull out the surface node information for coupling
    # ==========================================================================

    # using dictionaries to index with 1-based indexing for the bodies to match FUN3D
    num_nodes = {}
    x = {}
    y = {}
    z = {}
    for ibody in range(1, nbodies + 1):
        num_nodes[ibody] = flow.extract_surface_num(body=ibody)
        if num_nodes[ibody] > 0:
            x[ibody], y[ibody], z[ibody] = flow.extract_surface(
                num_nodes[ibody], body=ibody
            )

    # ==========================================================================
    # Time step loop
    # ==========================================================================
    for step in range(prior_steps + 1, prior_steps + 1 + nsteps):
        for ibody in range(1, nbodies + 1):
            if num_nodes[ibody] > 0:
                dx, dy, dz = get_deform_motion(step, x, y, z, ibody)
                flow.input_deformation(dx, dy, dz, body=ibody)

        flow.iterate()

        # running total of all the forces on all bodies on the local processor
        fx_proc_total = 0.0
        fy_proc_total = 0.0
        fz_proc_total = 0.0

        # sum of the forces on body i, across all processors
        fx_sum = {}
        fy_sum = {}
        fz_sum = {}

        for ibody in range(1, nbodies + 1):
            fx = 0.0
            fy = 0.0
            fz = 0.0
            if num_nodes[ibody] > 0:
                # forces on body i on the local processor. In the same order as x,y,z
                fx, fy, fz = flow.extract_forces(num_nodes[ibody], body=ibody)

                fx_proc_total += np.sum(fx)
                fy_proc_total += np.sum(fy)
                fz_proc_total += np.sum(fz)

            fx_sum[ibody] = comm.allreduce(np.sum(fx))
            fy_sum[ibody] = comm.allreduce(np.sum(fy))
            fz_sum[ibody] = comm.allreduce(np.sum(fz))

        # sum of all forces on all processors
        fx_total = comm.allreduce(fx_proc_total)
        fy_total = comm.allreduce(fy_proc_total)
        fz_total = comm.allreduce(fz_proc_total)

        if comm.Get_rank() == 0:
            for ibody in range(1, nbodies + 1):
                print(
                    "Total force on body ",
                    ibody,
                    " = ",
                    fx_sum[ibody],
                    fy_sum[ibody],
                    fz_sum[ibody],
                )
            print("Total on both bodies", fx_total, fy_total, fz_total)
    flow.post()


if __name__ == "__main__":
    main()
